/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_gridftp_server.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <grp.h>
#include <openssl/md5.h>
#include <string.h>
#include <netdb.h>
#include <zlib.h>
#include "Castor_limits.h"
#include "dpm_api.h"
#include "dpns_api.h"
#include "dsi_dpm.h"
#include "rfio_api.h"
#include "serrno.h"
#include "voms.h"

#define ERRBUFSIZE  1024
int Cdomainname (char *, int);
int Csec_isIdAService (const char *, const char *);
int isTrustedHost2 (char *, char *, char *, char *, char *);
int rfio_parse(char *, char **, char **);
extern int rfio_HsmIf_getipath(int, char *, char *);

gid_t dpm_gid;
uid_t dpm_uid;

/* For preserving the virtual ids */
gid_t Csec_gid;
uid_t Csec_uid;
/* For preserving the VOMS context */
gssapi_voms_ctx_t voms_ctx;

char data_interface[CA_MAXHOSTNAMELEN+1];
char dpnshost[CA_MAXHOSTNAMELEN+1];
char hostname[CA_MAXHOSTNAMELEN+1];
char localdomain[CA_MAXHOSTNAMELEN+1];
struct passwd *pw;

#include <syslog.h>
#include "gssapi.h"
#include "gssapi_openssl.h"
#include "globus_gsi_credential.h"

#ifdef USE_VOMS
#include "voms_apic.h"
#endif
static int _get_user_ca (X509 *px509_cred, STACK_OF(X509) *px509_chain, char *user_ca)
{
	X509 *cert;
	globus_gsi_cert_utils_cert_type_t cert_type;
	int i;

	cert = px509_cred;
	if (globus_gsi_cert_utils_get_cert_type(cert, &cert_type) != GLOBUS_SUCCESS)
		return (-1);
	if (cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_EEC ||
	    cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_CA) {
		X509_NAME_oneline(X509_get_issuer_name(cert), user_ca, CA_MAXCSECNAMELEN);
		return (0);
	}
	for (i = 0; i < sk_X509_num(px509_chain); i++) {
		cert = sk_X509_value (px509_chain, i);
		if (globus_gsi_cert_utils_get_cert_type(cert, &cert_type) != GLOBUS_SUCCESS)
			return (-1);
		if (cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_EEC ||
		    cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_CA) {
			X509_NAME_oneline(X509_get_issuer_name(cert), user_ca, CA_MAXCSECNAMELEN);
			return (0);
		}
	}
	return (-1);
}

/* this function extracts the VOMS extensions from user's credentials */
int get_userca_and_voms_creds(gssapi_voms_ctx_t *ctx, char *user_ca, gss_cred_id_t cred)
{
	X509 *px509_cred= NULL;
	STACK_OF(X509) *px509_chain = NULL;
#ifdef USE_VOMS
	int error= 0;
	struct vomsdata *vd= NULL;
	struct voms **volist;
#endif
	gss_cred_id_desc *       cred_desc = NULL;
	globus_gsi_cred_handle_t gsi_cred_handle;
	int ret = -1;

	/* cast to gss_cred_id_desc */
	if (cred == GSS_C_NO_CREDENTIAL) {
		return -1;
	}

	cred_desc = (gss_cred_id_desc *) cred;

	if (globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE) != GLOBUS_SUCCESS) {
		return -1;
	}

	/* Getting the X509 certificate */
	gsi_cred_handle = cred_desc->cred_handle;
	if (globus_gsi_cred_get_cert(gsi_cred_handle, &px509_cred) != GLOBUS_SUCCESS) {
		globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
		return -1;
	}

	/* Getting the certificate chain */
	if (globus_gsi_cred_get_cert_chain (gsi_cred_handle, &px509_chain) != GLOBUS_SUCCESS) {
		X509_free (px509_cred);
		(void)globus_module_deactivate (GLOBUS_GSI_CREDENTIAL_MODULE);
		return -1;
	}

	if (_get_user_ca (px509_cred, px509_chain, user_ca) < 0)
		goto leave;

	/* No need for the globus module anymore, the rest are calls to VOMS */
	(void)globus_module_deactivate (GLOBUS_GSI_CREDENTIAL_MODULE);

#if defined(USE_VOMS)

	if ((vd = VOMS_Init (NULL, NULL)) == NULL) {
		goto leave;
	}

	if (VOMS_Retrieve (px509_cred, px509_chain, RECURSE_CHAIN, vd, &error) == 0) {
		if (error == VERR_NOEXT) /* Non voms proxies will fall into this error */
		{
			/* Non voms proxies go here: set the number of fqans to zero. Not an error */
			ctx->nbfqan = 0;
			ret = 0;
		}
		else
		{
			char buffer[ERRBUFSIZE];
			VOMS_ErrorMessage(vd, error, buffer, ERRBUFSIZE);
			/* if strncat fails, it will not be pretty, but the  *
			 * error message is there, so its failure is ignored */
			strncat(buffer, "\n", ERRBUFSIZE);
			syslog(LOG_INFO, buffer);
		}
		VOMS_Destroy (vd);
		goto leave;
	}
	volist = vd->data;

	if (volist !=NULL) {
		int i = 0;
		int nbfqan;

		/* Copying the voname */
		if ((*volist)->voname != NULL) {
			ctx->voname = strdup((*volist)->voname);
		}

		/* Counting the fqans before allocating the array */
		while( volist[0]->fqan[i] != NULL) {
			i++;
		}
		nbfqan = i;

		if (nbfqan > 0) {
			ctx->fqan = malloc(sizeof(char *) * (i+1));
			if (ctx->fqan != NULL) {
				for (i=0; i<nbfqan; i++) {
					ctx->fqan[i] = strdup( volist[0]->fqan[i]);
				}
				ctx->fqan[nbfqan] = NULL;
				ctx->nbfqan = nbfqan;
			}
		} /* if (nbfqan > 0) */
	}
	VOMS_Destroy (vd);

#endif /* USE_VOMS */

	ret = 0;

leave:
	if (px509_cred) X509_free (px509_cred);
	if (px509_chain) sk_X509_pop_free(px509_chain,X509_free);

	return ret;
}

static
globus_version_t local_version =
{
	0, /* major version number */
	2, /* minor version number */
	1174988050,
	0 /* branch ID */
};

/*
 *  utility function to make errors
 */
static
globus_result_t
globus_l_gfs_rfio_make_error(
		const char *                        msg)
{
	char *                              err_str;
	globus_result_t                     result;
	GlobusGFSName(globus_l_gfs_dpm_make_error);

	err_str = globus_common_create_string("%s error: %s",
			msg,  rfio_serror());
	result = GlobusGFSErrorGeneric(err_str);
	free(err_str);
	return result;
}

/* abort a DPM file being written by the ftpd, given the fd of the file */
static 
int abort_dpm_fd (int fd) {
	char *p,name[CA_MAXSFNLEN+1], r_token[CA_MAXDPMTOKENLEN+1];
	struct dpm_filestatus *filestatuses = NULL;
	int rc,nbreplies = 0;

	rc = rfio_HsmIf_getipath (fd, name, r_token);
	if (rc<0) return (rc);

	p = name;
	rc = dpm_abortfiles (r_token, 1, &p, &nbreplies, &filestatuses);

	dpm_free_filest (nbreplies, filestatuses);
        return (rc);
}

/* change the pathname from the format like:
/hostname:/data/...
into hostname:/data
 */

char * changepath(char * pathname_in)
{
	char *pathname;
	const char *p,*p2;

	while (*pathname_in == '/' && *(pathname_in+1) == '/')
		pathname_in++;	/* reduce leading multiple slashes to one */

	/* first colon-slash */
	p = strstr(pathname_in, ":/");
   
	/* first slash after any leading slashes */
	p2 = strchr((*pathname_in == '/') ? (pathname_in+1) : pathname_in, '/');
        
	/* treat it as a hostname:/path if there is no slash before the colon-slash */
	if (p != NULL && p2>p) {
		p = pathname_in;
		if (*p == '/')
			p++;
		pathname = strdup(p);
	}
	else
		pathname = strdup(pathname_in);
	return (pathname);
}

int initdpm_client(globus_gfs_session_info_t *session_info, char *errbuf)
{
	struct addrinfo *aitop;
	char clienthost[CA_MAXHOSTNAMELEN+7];
	struct passwd *dpm_pwd;
	//char *func = "intdpm_client";
	int gaierrno;
	struct addrinfo hints;
	char *identity = session_info->subject;
	char *p;
	char user_ca[CA_MAXCSECNAMELEN+1];
	char *username;

	dpm_pwd = getpwnam("dpmmgr");
	if (!dpm_pwd) {
		syslog(LOG_ERR, "dpmmgr account is not defined in passwd file\n");
		strcpy(errbuf, "dpmmgr account is not defined in passwd file");
		return -1;
	}
	dpm_gid = dpm_pwd->pw_gid;
	dpm_uid = dpm_pwd->pw_uid;

	if (get_userca_and_voms_creds(&voms_ctx, user_ca, session_info->del_cred) == -1) {
		syslog(LOG_INFO, "user_ca/VOMS processing returned error\n");
		strcpy(errbuf, "user_ca/VOMS error when processing cert");
		return -1;
	}

	if (gethostname(hostname, CA_MAXHOSTNAMELEN+1)) {
		syslog (LOG_ERR, "Unable to get local host name\n");
		strcpy (errbuf, "Unable to get local host name");
		return -1;
	}
	memset(&hints, '\0', sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags |= AI_CANONNAME;
	gaierrno = getaddrinfo(hostname, NULL, &hints, &aitop);
	if (gaierrno != 0)
		aitop = NULL;
	if (aitop && aitop->ai_canonname) {
		struct addrinfo *res;
		memset(&hints, '\0', sizeof(hints));
		hints.ai_family = PF_UNSPEC;
		hints.ai_flags |= AI_NUMERICHOST;
		gaierrno = getaddrinfo(aitop->ai_canonname, NULL, &hints, &res);
		if (gaierrno == 0)
			freeaddrinfo(res);
		else if (strlen(aitop->ai_canonname)<=CA_MAXHOSTNAMELEN)
			strcpy(hostname, aitop->ai_canonname);
	}
	if (aitop)
		freeaddrinfo(aitop);

	if (Cdomainname (localdomain, sizeof(localdomain)) < 0) {
		syslog (LOG_ERR, "Unable to get local domain name\n");
		strcpy (errbuf, "Unable to get local domain name");
		return -1;
	}

	/* if the hostname is not fully qualified then add on the localdomain */
	if (strchr(hostname, '.') == NULL &&
	    strlen(hostname) + strlen(localdomain) < CA_MAXHOSTNAMELEN) {
		strcat(hostname, ".");
		strcat(hostname, localdomain);
	}

	globus_gridftp_server_get_config_string (NULL, &p);
	if (p) {
		strcpy (data_interface, p);
		if (strchr (data_interface, '.') == NULL &&
		    strlen(data_interface) + strlen(localdomain) < CA_MAXHOSTNAMELEN) {
			strcat (data_interface, ".");
			strcat (data_interface, localdomain);
		}
	} else
		strcpy (data_interface, hostname);

	if ((p = getenv ("DPNS_HOST")))
		strcpy (dpnshost, p);

	if (session_info->host_id) {
		strcpy (clienthost, session_info->host_id);
		p = strrchr (clienthost, ':');
		if (p && (*clienthost != '[' || *(p-1) == ']'))
			*p = '\0';
	} else
		clienthost[0] = '\0';
	globus_gfs_log_message(GLOBUS_GFS_LOG_INFO, "request by %s from %s\n", identity, clienthost);
	if (Csec_isIdAService ("GSI", identity) >= 0 &&
	    isTrustedHost2 (clienthost, hostname, localdomain, "RFIOD", "TRUST")) {
		Csec_uid = 0;
		Csec_gid = 0;
		username = "root";
	} else {
#ifdef VIRTUAL_ID	
		if (Cns_getidmapc (identity, user_ca, voms_ctx.nbfqan > 1 ? 1 : voms_ctx.nbfqan,
		    (const char **)voms_ctx.fqan, &Csec_uid, &Csec_gid)) {
			syslog(LOG_INFO, "getidmap() returned error\n");
			strcpy(errbuf, "Could not get virtual id!");
			return -1;
		}
#endif  /* VIRTUAL_ID */
		if (globus_gss_assist_gridmap (identity, &username) != 0) {
			syslog(LOG_INFO, "No local mapping\n");
			strcpy(errbuf, "No local mapping");
			return -1;
		}
	}
	if (session_info->username)
		globus_free (session_info->username);
	session_info->username = strdup (username);
	pw = getpwnam(username);
	if (!pw) {
		syslog(LOG_INFO, "No local mapping\n");
		strcpy(errbuf, "No local mapping");
		return -1;
	}
	setgroups(0, NULL);
	setegid(pw->pw_gid);
	seteuid(pw->pw_uid);
	return 0;
}



/*************************************************************************
 *  start
 *  -----
 *  This function is called when a new session is initialized, ie a user 
 *  connectes to the server.  This hook gives the dsi an oppertunity to
 *  set internal state that will be threaded through to all other
 *  function calls associated with this session.  And an oppertunity to
 *  reject the user.
 *
 *  finished_info.info.session.session_arg should be set to an DSI
 *  defined data structure.  This pointer will be passed as the void *
 *  user_arg parameter to all other interface functions.
 * 
 *  NOTE: at nice wrapper function should exist that hides the details 
 *        of the finished_info structure, but it currently does not.  
 *        The DSI developer should jsut follow this template for now
 ************************************************************************/
static
void
globus_l_gfs_dpm_start(
		globus_gfs_operation_t              op,
		globus_gfs_session_info_t *         session_info)
{
	globus_l_gfs_dpm_handle_t *		dpm_handle;
	globus_gfs_finished_info_t          finished_info;
	globus_result_t                     result; 
	char errbuf[256];
	//char *				func = "dpm_start";

	GlobusGFSName(globus_l_gfs_dpm_start);

	dpm_handle = (globus_l_gfs_dpm_handle_t *)
		globus_malloc(sizeof(globus_l_gfs_dpm_handle_t));

        // Initialize structs (with success)
	globus_mutex_init(&dpm_handle->mutex, NULL); 
	memset(&finished_info, '\0', sizeof(globus_gfs_finished_info_t));
	finished_info.type = GLOBUS_GFS_OP_SESSION_START;
	finished_info.info.session.session_arg = dpm_handle;
	finished_info.info.session.username = session_info->username;
	finished_info.info.session.home_dir = "/";
        result = GLOBUS_SUCCESS;

        // Set result appropriately if error
	if(session_info->del_cred == NULL) {
                strcpy(errbuf, "must be GSSAPI auth for DPM");
		result = GlobusGFSErrorGeneric(errbuf);
	}
	if (initdpm_client(session_info, errbuf) < 0) {
		result = GlobusGFSErrorGeneric(errbuf);
	}

        // And finish up
        finished_info.result = result;
        finished_info.msg = errbuf; 

	dpm_handle->fd = -1;
	dpm_handle->subject = strdup(session_info->subject);

	globus_gridftp_server_operation_finished(op, result, &finished_info);
	return ;
}

/*************************************************************************
 *  destroy
 *  -------
 *  This is called when a session ends, ie client quits or disconnects.
 *  The dsi should clean up all memory they associated wit the session
 *  here. 
 ************************************************************************/
static
void
globus_l_gfs_dpm_destroy(
		void *                              user_arg)
{
	globus_l_gfs_dpm_handle_t *		dpm_handle;

	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;
	globus_mutex_destroy(&dpm_handle->mutex);
	free(dpm_handle->subject);
	globus_free(dpm_handle);
}

/* fill the statbuf into globus_gfs_stat_t */
void fill_stat_array(globus_gfs_stat_t * filestat, struct stat64 statbuf, char *name)
{
	filestat->mode = statbuf.st_mode;;
	filestat->nlink = statbuf.st_nlink;
	filestat->uid = statbuf.st_uid;
	filestat->gid = statbuf.st_gid;
	filestat->size = statbuf.st_size;

	filestat->mtime = statbuf.st_mtime;
	filestat->atime = statbuf.st_atime;
	filestat->ctime = statbuf.st_ctime;

	filestat->dev = statbuf.st_dev;
	filestat->ino = statbuf.st_ino;
	filestat->name = strdup(name);
}


/*************************************************************************
 *  stat
 *  ----
 *  This interface function is called whenever the server needs 
 *  information about a given file or resource.  It is called then an
 *  LIST is sent by the client, when the server needs to verify that 
 *  a file exists and has the proper permissions, etc.
 ************************************************************************/
static
void
globus_l_gfs_dpm_stat(
		globus_gfs_operation_t              op,
		globus_gfs_stat_info_t *            stat_info,
		void *                              user_arg)
{
	int                                 stat_count;
	int                                 status;
	int                                 nbsub;
	int                                 i;
	char                                filename[CA_MAXPATHLEN+1];
	char *				pathname;
	struct stat64                       statbuf;
	globus_gfs_stat_t                   *stat_array;
	globus_l_gfs_dpm_handle_t *        	dpm_handle;
	DIR *                               dp;
	struct dirent64 *                   dirp;
	globus_result_t                     result;
	//char *				func = "dpm_stat";

	GlobusGFSName(globus_l_gfs_dpm_stat);

	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;

	pathname = changepath(stat_info->pathname);

	//status = rfio_stat64(pathname, &statbuf);
	status = dpm_handle_stat(pathname, &statbuf, dpm_handle->subject);

	if (status != 0)
		goto out;
	if (stat_info->file_only)
		goto filestat;
	if (S_ISDIR(statbuf.st_mode)) {//dir
		stat_count = 0;
		nbsub = 0;
		dp = rfio_opendir(pathname);
		if (!dp) {
			goto filestat;
		}
		while ((dirp = rfio_readdir64(dp)) != NULL) {
			stat_count++;
		}
		rfio_rewinddir(dp);
		if (stat_count == 0) {
			rfio_closedir(dp);
			goto filestat;
		}

		stat_array = (globus_gfs_stat_t *)globus_calloc(stat_count, sizeof(globus_gfs_stat_t));
		while ((dirp = rfio_readdir64(dp)) != NULL) {
			sprintf(filename, "%s/%s", pathname, dirp->d_name);
			//status = rfio_stat64(filename, &statbuf);
			status = dpm_handle_stat(filename, &statbuf, dpm_handle->subject);
			if (status == 0) {
				fill_stat_array(&(stat_array[nbsub]), statbuf, dirp->d_name);
				nbsub++;
			}
		}
		rfio_closedir(dp);
		globus_gridftp_server_finished_stat(op, GLOBUS_SUCCESS, stat_array, nbsub);
		for (i = 0; i < nbsub; i++)
			free(stat_array[i].name);
		globus_free(stat_array);
		free(pathname);
		return;
	}
filestat:
	stat_array = (globus_gfs_stat_t *)globus_calloc(1, sizeof(globus_gfs_stat_t));
	stat_count = 1;
	fill_stat_array(&(stat_array[0]), statbuf, pathname);
	globus_gridftp_server_finished_stat(op, GLOBUS_SUCCESS, stat_array, stat_count);
	globus_free(stat_array);
	free(pathname);
	return;
out:
	free(pathname);
	result = globus_l_gfs_rfio_make_error("stat");
	globus_gridftp_server_finished_stat(op, result, NULL, 0);
	return;
}

/*************************************************************************
 *  command
 *  -------
 *  This interface function is called when the client sends a 'command'.
 *  commands are such things as mkdir, remdir, delete.  The complete
 *  enumeration is below.
 *
 *  To determine which command is being requested look at:
 *      cmd_info->command
 *
 *      GLOBUS_GFS_CMD_MKD = 1,
 *      GLOBUS_GFS_CMD_RMD,
 *      GLOBUS_GFS_CMD_DELE,
 *      GLOBUS_GFS_CMD_RNTO,
 *      GLOBUS_GFS_CMD_RNFR,
 *      GLOBUS_GFS_CMD_CKSM,
 *      GLOBUS_GFS_CMD_SITE_CHMOD,
 *      GLOBUS_GFS_CMD_SITE_DSI
 ************************************************************************/

static
globus_result_t
compute_cksm(
		char *                              pathname,
		const char *                        algorithm,
		globus_off_t                        offset,
		globus_off_t                        length,
		char *                              checksum,
		char *                              subject)
{
	globus_byte_t * buffer;
	unsigned long	ckSum;
	int		ckTyp;
	static char   * ckTyp_DPM[] = {"MD", "AD", "CS"};
	int		fd;
	struct Cns_fileid file_uniqueid;
	struct Cns_filestatg filestatg;
	int		i;
	unsigned char	md[MD5_DIGEST_LENGTH];
	MD5_CTX		md5_ctx;
	globus_off_t	nbread;
	char *		p;
	globus_off_t	read_left;
	globus_off_t	read_length;
	globus_result_t	result;
	GlobusGFSName(compute_cksm);

	if (strcmp (algorithm, "MD5") == 0 || strcmp (algorithm, "md5") == 0)
		ckTyp = 0;
	else if (strcmp (algorithm, "ADLER32") == 0 || strcmp (algorithm, "adler32") == 0)
		ckTyp = 1;
	else if (strcmp (algorithm, "CRC32") == 0 || strcmp (algorithm, "crc32") == 0)
		ckTyp = 2;
	else {
		serrno = SEOPNOTSUP;
		result = globus_l_gfs_rfio_make_error("CKSM");
		return result;
	}
	fd = dpm_handle_open(pathname, O_RDONLY, 0644, subject);
	if (fd < 0) {
		result = globus_l_gfs_rfio_make_error("CKSM: open");
		return result;
	}
	if (length > 0) {
		read_left = length;
		read_length = (length > 1024 * 1024) ? 1024 * 1024 : length;
	} else
		read_length = 1024 * 1024;
	if (offset > 0) {
		if (rfio_lseek64(fd, offset, SEEK_SET) != offset) {
			result = globus_l_gfs_rfio_make_error("CKSM: seek");
			goto error;
		}
	}
	buffer = globus_malloc(read_length);
	if(buffer == NULL) {
		result = GlobusGFSErrorGeneric("CKSM: malloc failed");
		goto error;
	}
	if (ckTyp == 0)
		MD5_Init(&md5_ctx);
	else if (ckTyp == 1)
		ckSum = adler32(0L, Z_NULL, 0);
	else
		ckSum = crc32(0L, Z_NULL, 0);
	while ((nbread = rfio_read(fd, buffer, read_length)) > 0) {
		if (ckTyp == 0)
			MD5_Update(&md5_ctx, buffer, nbread);
		else if (ckTyp == 1)
			ckSum = adler32(ckSum, buffer, nbread);
		else
			ckSum = crc32(ckSum, buffer, nbread);
		if (length > 0) {
			read_left -= nbread;
			if (read_left <= 0) {
				nbread = 0;
				break;
			}
			read_length = (read_left > 1024 * 1024) ? 1024 * 1024 : read_left;
		}
	}
	if(nbread < 0) {	/* error */
		result = globus_l_gfs_rfio_make_error("CKSM: read");
		goto free_buffer;
	}
	if (ckTyp == 0) {
		MD5_Final(md, &md5_ctx);
		p = checksum;
		for (i = 0; i < MD5_DIGEST_LENGTH; i++) {
			sprintf(p, "%02x", md[i]);
			p += 2;
		}
		*p = '\0';
	} else
		sprintf(checksum, "%08x", (unsigned int) ckSum);

	/* store the computed checksum in DPNS if no checksum stored yet. */
	if (dpm_handle_statc (pathname, &filestatg, subject) == 0) {
		if ((length <= 0 || length == filestatg.filesize) && offset == 0 &&
		    ! *filestatg.csumtype) {
			strcpy (file_uniqueid.server, dpnshost);
			file_uniqueid.fileid = filestatg.fileid;
			if (Cns_setfsizec (NULL, &file_uniqueid, filestatg.filesize,
			    ckTyp_DPM[ckTyp], checksum) < 0) {
				result = globus_l_gfs_rfio_make_error("CKSM: setfsizec");
				goto free_buffer;
			}
		}
	}

	result = GLOBUS_SUCCESS;
free_buffer:
	globus_free(buffer);
error:
	rfio_close(fd);
	return result;
}

static
void
globus_l_gfs_dpm_command(
		globus_gfs_operation_t              op,
		globus_gfs_command_info_t *         cmd_info,
		void *                              user_arg)
{
	int                                 status = -1;
	globus_l_gfs_dpm_handle_t *         dpm_handle;
	char *                              filename;
	char *                              frm_pathname;
	char *                              host;
	char                                md5sum[MD5_DIGEST_LENGTH*2+1];
	char *                              ops;
	int                                 parse_rc;
	char *                              pathname;
	globus_result_t                     result;
	GlobusGFSName(globus_l_gfs_dpm_command);

	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;

	pathname = changepath(cmd_info->pathname);
	if(pathname == NULL) {
		result = GlobusGFSErrorGeneric("RFIO: strdup failed");
		goto alloc_error;
	}

	if (Csec_uid &&
	    (parse_rc = rfio_parse(pathname, &host, &filename)) >= 0) {
		if (parse_rc) {		/* remote file */
			rfio_client_setAuthorizationId(Csec_uid, Csec_gid,
			    "GSI", dpm_handle->subject);
			if (voms_ctx.voname && voms_ctx.fqan)
				rfio_client_setVOMS_data(voms_ctx.voname,
				    voms_ctx.fqan, voms_ctx.nbfqan);
		} else if (host) {	/* DPM managed file */
			if (cmd_info->command == GLOBUS_GFS_CMD_DELE) {
				dpm_client_setAuthorizationId(Csec_uid, Csec_gid,
				    "GSI", dpm_handle->subject);
				if (voms_ctx.voname && voms_ctx.fqan)
					dpm_client_setVOMS_data(voms_ctx.voname,
					    voms_ctx.fqan, voms_ctx.nbfqan);
			} else {
				Cns_client_setAuthorizationId(Csec_uid, Csec_gid,
				    "GSI", dpm_handle->subject);
				if (voms_ctx.voname && voms_ctx.fqan)
					Cns_client_setVOMS_data(voms_ctx.voname,
					    voms_ctx.fqan, voms_ctx.nbfqan);
			}
		}
	}
	switch(cmd_info->command)
	{
		case GLOBUS_GFS_CMD_MKD:
			ops = "mkdir";
			status = rfio_mkdir(pathname, 0755);
			break;

		case GLOBUS_GFS_CMD_RMD:
			ops = "rmdir";
			status = rfio_rmdir(pathname);
			break;

		case GLOBUS_GFS_CMD_DELE:
			ops = "unlink";
			status = rfio_unlink(pathname);
			break;

		case GLOBUS_GFS_CMD_RNTO:
			ops = "rename";
			frm_pathname = changepath(cmd_info->rnfr_pathname);
			if(frm_pathname == NULL) {
				result = GlobusGFSErrorGeneric("RFIO: strdup failed");
				goto error;
			}
			status = rfio_rename(frm_pathname, pathname);
			free(frm_pathname);
			break;

		case GLOBUS_GFS_CMD_CKSM:
			result = compute_cksm(pathname, cmd_info->cksm_alg,
			    cmd_info->cksm_offset, cmd_info->cksm_length,
			    md5sum, dpm_handle->subject);
			if (result != GLOBUS_SUCCESS)
				goto error;
			globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, md5sum);
			free(pathname);
			return;
		case GLOBUS_GFS_CMD_SITE_CHMOD:
			ops = "chmod";
			status = rfio_chmod(pathname, cmd_info->chmod_mode);
			break;
		default:
			ops = "";
			break;
	}

	if(status != 0) {
		result = globus_l_gfs_rfio_make_error(ops);
		goto error;
	}

	globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, GLOBUS_NULL);

	free(pathname);
	return;

error:
	free(pathname);
alloc_error:
	globus_gridftp_server_finished_command(op, result, NULL);

}

/*************************************************************************
 *  recv
 *  ----
 *  This interface function is called when the client requests that a
 *  file be transfered to the server.
 *
 *  To receive a file the following functions will be used in roughly
 *  the presented order.  They are doced in more detail with the
 *  gridftp server documentation.
 *
 *      globus_gridftp_server_begin_transfer();
 *      globus_gridftp_server_register_read();
 *      globus_gridftp_server_finished_transfer();
 *
 ************************************************************************/
/* receive from client */
static
void
globus_l_gfs_rfio_net_read_cb(
		globus_gfs_operation_t              op,
		globus_result_t                     result,
		globus_byte_t *                     buffer,
		globus_size_t                       nbytes,
		globus_off_t                        offset,
		globus_bool_t                       eof,
		void *                              user_arg)
{
	globus_off_t                        start_offset;
	globus_l_gfs_dpm_handle_t *         dpm_handle;
	globus_size_t                       bytes_written;
	//char *				func = "globus_l_gfs_rfio_net_read_cb";                                                                                                             
	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;

	globus_mutex_lock(&dpm_handle->mutex);
	{
		if(eof)
			dpm_handle->done = GLOBUS_TRUE;
		dpm_handle->outstanding--;
		if(result != GLOBUS_SUCCESS) {
			dpm_handle->cached_res = result;
			dpm_handle->done = GLOBUS_TRUE;
		}
		else if(nbytes > 0) {
			start_offset = rfio_lseek64(dpm_handle->fd, offset, SEEK_SET);
			if(start_offset != offset) {
				dpm_handle->cached_res = globus_l_gfs_rfio_make_error("seek");
				dpm_handle->done = GLOBUS_TRUE;
			}
			else {
				bytes_written = rfio_write(dpm_handle->fd, buffer, nbytes);
				if(bytes_written == -1 || bytes_written < nbytes) {
					if (bytes_written != -1) serrno = ENOSPC;
					dpm_handle->cached_res = globus_l_gfs_rfio_make_error("write");
					dpm_handle->done = GLOBUS_TRUE;
				} else
					globus_gridftp_server_update_bytes_written(op, offset, nbytes);
			}
		}

		globus_free(buffer);
		/* if not done just register the next one */
		if(!dpm_handle->done)
			globus_l_gfs_dpm_read_from_net(dpm_handle);
		/* if done and there are no outstanding callbacks finish */
		else if(dpm_handle->outstanding == 0)
		{
			if (dpm_handle->cached_res != GLOBUS_SUCCESS) {
				(void) abort_dpm_fd (dpm_handle->fd);
			}
			rfio_close(dpm_handle->fd);
			globus_gridftp_server_finished_transfer(op, dpm_handle->cached_res);
		}
	}
	globus_mutex_unlock(&dpm_handle->mutex);
}


static
void
globus_l_gfs_dpm_read_from_net(
		globus_l_gfs_dpm_handle_t *         dpm_handle)
{
	globus_byte_t *                     buffer;
	globus_result_t                     result;
	GlobusGFSName(globus_l_gfs_dpm_read_from_net);

	/* in the read case tis number will vary */
	globus_gridftp_server_get_optimal_concurrency(
			dpm_handle->op, &dpm_handle->optimal_count);

	while(dpm_handle->outstanding < dpm_handle->optimal_count) {
		buffer = globus_malloc(dpm_handle->block_size);
		if(buffer == NULL) {
			result = GlobusGFSErrorGeneric("malloc failed");
			goto error;
		}
		result = globus_gridftp_server_register_read(
				dpm_handle->op,
				buffer,
				dpm_handle->block_size,
				globus_l_gfs_rfio_net_read_cb,
				dpm_handle);
		if(result != GLOBUS_SUCCESS)  {
			goto alloc_error;
		}
		dpm_handle->outstanding++;
	}

	return;

alloc_error:
	globus_free(buffer);
error:
	dpm_handle->cached_res = result;
	dpm_handle->done = GLOBUS_TRUE;
	if(dpm_handle->outstanding == 0)
	{
		if (dpm_handle->cached_res != GLOBUS_SUCCESS) {
			(void) abort_dpm_fd (dpm_handle->fd);
		}
		rfio_close(dpm_handle->fd);
		globus_gridftp_server_finished_transfer(
				dpm_handle->op, dpm_handle->cached_res);
	}
}

static
void
globus_l_gfs_dpm_recv(
		globus_gfs_operation_t              op,
		globus_gfs_transfer_info_t *        transfer_info,
		void *                              user_arg)
{
	char *                              pathname;
	int                                 flags;
	globus_result_t                     result;
	globus_l_gfs_dpm_handle_t *         dpm_handle;

	GlobusGFSName(globus_l_gfs_dpm_recv);

	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;

	pathname = changepath(transfer_info->pathname);
	if(pathname == NULL)  {
		result = GlobusGFSErrorGeneric("RFIO: strdup failed");
		goto alloc_error;
	}

	/* try to open */
	flags = O_WRONLY | O_CREAT;
	if(transfer_info->truncate) {
		flags |= O_TRUNC;
	}
	dpm_handle->fd = dpm_handle_open(pathname, flags, 0644, dpm_handle->subject);
	if(dpm_handle->fd < 0) {
		result = globus_l_gfs_rfio_make_error("open/create");
		goto error;
	}

	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;

	/* reset all the needed variables in the handle */
	dpm_handle->cached_res = GLOBUS_SUCCESS;
	dpm_handle->outstanding = 0;
	dpm_handle->done = GLOBUS_FALSE;
	dpm_handle->blk_length = 0;
	dpm_handle->blk_offset = 0;
	dpm_handle->op = op;
	globus_gridftp_server_get_block_size(
			op, &dpm_handle->block_size);

	globus_gridftp_server_begin_transfer(op, 0, dpm_handle);

	globus_mutex_lock(&dpm_handle->mutex);
	{
		globus_l_gfs_dpm_read_from_net(dpm_handle);
	}
	globus_mutex_unlock(&dpm_handle->mutex);
	return;
error:
	globus_free(pathname);
alloc_error:
	globus_gridftp_server_finished_transfer(op, result);


}

/*************************************************************************
 *  send
 *  ----
 *  This interface function is called when the client requests to receive
 *  a file from the server.
 *
 *  To send a file to the client the following functions will be used in roughly
 *  the presented order.  They are doced in more detail with the
 *  gridftp server documentation.
 *
 *      globus_gridftp_server_begin_transfer();
 *      globus_gridftp_server_register_write();
 *      globus_gridftp_server_finished_transfer();
 *
 ************************************************************************/
static
void
globus_l_gfs_net_write_cb(
		globus_gfs_operation_t              op,
		globus_result_t                     result,
		globus_byte_t *                     buffer,
		globus_size_t                       nbytes,
		void *                              user_arg)
{
	globus_l_gfs_dpm_handle_t *         dpm_handle;

	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;
	globus_free(buffer);
	globus_mutex_lock(&dpm_handle->mutex);
	{

		dpm_handle->outstanding--;
		if(result != GLOBUS_SUCCESS) {
			dpm_handle->cached_res = result;
			dpm_handle->done = GLOBUS_TRUE;
		}
		if(!dpm_handle->done)  {
			globus_l_gfs_dpm_send_next_to_client(dpm_handle);
		}
		else if (dpm_handle->outstanding == 0) {
			rfio_close(dpm_handle->fd);
			globus_gridftp_server_finished_transfer(op, dpm_handle->cached_res);        
		}
	}
	globus_mutex_unlock(&dpm_handle->mutex);
}


static
globus_bool_t
globus_l_gfs_dpm_send_next_to_client(
		globus_l_gfs_dpm_handle_t *         dpm_handle)
{
	globus_result_t                     result;
	globus_result_t                     res;
	globus_off_t                        read_length;
	globus_off_t                        nbread;
	globus_off_t                        start_offset;
	globus_byte_t *                     buffer;
	//char *				func = "globus_l_gfs_dpm_send_next_to_client";

	GlobusGFSName(globus_l_gfs_dpm_send_next_to_client);

	if (dpm_handle->blk_length == 0) {
		globus_gridftp_server_get_read_range(
				dpm_handle->op,
				&dpm_handle->blk_offset,
				&dpm_handle->blk_length);
		if(dpm_handle->blk_length == 0) {
			result = GLOBUS_SUCCESS;
			goto error;
		}
	}
	if(dpm_handle->blk_length == -1 ||
			dpm_handle->blk_length > dpm_handle->block_size) {
		read_length = dpm_handle->block_size;
	}
	else {
		read_length = dpm_handle->blk_length;
	}

	start_offset = rfio_lseek64(dpm_handle->fd, dpm_handle->blk_offset, SEEK_SET);
	/* verify that it worked */
	if(start_offset != dpm_handle->blk_offset) {
		result = globus_l_gfs_rfio_make_error("seek");
		goto error;
	}
	buffer = globus_malloc(read_length);
	if(buffer == NULL) {
		result = GlobusGFSErrorGeneric("RFIO: malloc failed");
		goto error;
	}
	nbread = rfio_read(dpm_handle->fd, buffer, read_length);
	if(nbread == 0) {	/* eof */
		result = GLOBUS_SUCCESS;
		goto alloc_error;
	}
	if(nbread < 0) {	/* error */
		result = globus_l_gfs_rfio_make_error("read");
		goto alloc_error;
	}
	read_length = nbread;      

	dpm_handle->blk_offset += read_length;
	if(dpm_handle->blk_length != -1)  {
		dpm_handle->blk_length -= read_length;
	}

	res = globus_gridftp_server_register_write(
			dpm_handle->op, buffer, read_length, start_offset, -1,
			globus_l_gfs_net_write_cb, dpm_handle);
	if(res != GLOBUS_SUCCESS) {
		goto alloc_error;
	}
	dpm_handle->outstanding++;

	return GLOBUS_FALSE;

alloc_error:
	globus_free(buffer);
error:
	rfio_close(dpm_handle->fd);
	dpm_handle->cached_res = result;
	dpm_handle->done = GLOBUS_TRUE;
	if (dpm_handle->outstanding == 0) {
		globus_gridftp_server_finished_transfer(
				dpm_handle->op, dpm_handle->cached_res);
	}
	return dpm_handle->done;
}

static
void
globus_l_gfs_dpm_send(
		globus_gfs_operation_t              op,
		globus_gfs_transfer_info_t *        transfer_info,
		void *                              user_arg)
{
	globus_l_gfs_dpm_handle_t *       	dpm_handle;
	globus_bool_t                       done = GLOBUS_FALSE;
	char *                              pathname;
	globus_result_t                     result;
	int 				i;

	GlobusGFSName(globus_l_gfs_dpm_send);
	dpm_handle = (globus_l_gfs_dpm_handle_t *) user_arg;

	pathname = changepath(transfer_info->pathname);
	if (pathname == NULL) {
		result = GlobusGFSErrorGeneric("RFIO: strdup failed");
		goto alloc_error;
	}
	dpm_handle->fd = dpm_handle_open(pathname, O_RDONLY, 0644, dpm_handle->subject);
	if(dpm_handle->fd < 0) {
		result = globus_l_gfs_rfio_make_error("open");
		goto error;
	}
	/* reset all the needed variables in the handle */
	dpm_handle->cached_res = GLOBUS_SUCCESS;
	dpm_handle->outstanding = 0;
	dpm_handle->done = GLOBUS_FALSE;
	dpm_handle->blk_length = 0;
	dpm_handle->blk_offset = 0;
	dpm_handle->op = op;

	globus_gridftp_server_get_optimal_concurrency(
			op, &dpm_handle->optimal_count);
	globus_gridftp_server_get_block_size(
			op, &dpm_handle->block_size);
	dpm_handle->optimal_count = 1;

	globus_gridftp_server_begin_transfer(op, 0, dpm_handle);
	globus_mutex_lock(&dpm_handle->mutex);
	{
		for(i = 0; i < dpm_handle->optimal_count && !done; i++) {
			done = globus_l_gfs_dpm_send_next_to_client(dpm_handle);
		}
	}
	globus_mutex_unlock(&dpm_handle->mutex);
	globus_free(pathname);
	return;


error:
	globus_free(pathname);
alloc_error:
	globus_gridftp_server_finished_transfer(op, result);

}

static
int
globus_l_gfs_dpm_activate(void);

static
int
globus_l_gfs_dpm_deactivate(void);

/*
 *  no need to change this
 */
static globus_gfs_storage_iface_t       globus_l_gfs_dpm_dsi_iface = 
{
	GLOBUS_GFS_DSI_DESCRIPTOR_BLOCKING | GLOBUS_GFS_DSI_DESCRIPTOR_SENDER,
	globus_l_gfs_dpm_start,
	globus_l_gfs_dpm_destroy,
	NULL, /* list */
	globus_l_gfs_dpm_send,
	globus_l_gfs_dpm_recv,
	NULL, /* trev */
	NULL, /* active */
	NULL, /* passive */
	NULL, /* data destroy */
	globus_l_gfs_dpm_command, 
	globus_l_gfs_dpm_stat,
	NULL,
	NULL
};

/*
 *  no need to change this
 */
GlobusExtensionDefineModule(globus_gridftp_server_dpm) =
{
	"globus_gridftp_server_dpm",
	globus_l_gfs_dpm_activate,
	globus_l_gfs_dpm_deactivate,
	NULL,
	NULL,
	&local_version
};

/*
 *  no need to change this
 */
static
	int
globus_l_gfs_dpm_activate(void)
{
	globus_extension_registry_add(
			GLOBUS_GFS_DSI_REGISTRY,
			"dpm",
			GlobusExtensionMyModule(globus_gridftp_server_dpm),
			&globus_l_gfs_dpm_dsi_iface);

	return 0;
}

/*
 *  no need to change this
 */
static
	int
globus_l_gfs_dpm_deactivate(void)
{
	globus_extension_registry_remove(
			GLOBUS_GFS_DSI_REGISTRY, "dpm");

	return 0;
}
